"""
Copyright (C) 2012 Martin Blech and individual contributors.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from xml.parsers import expat
from xml.sax.saxutils import XMLGenerator
from xml.sax.xmlreader import AttributesImpl
try:  # pragma no cover
    from cStringIO import StringIO
except ImportError:  # pragma no cover
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO
try:  # pragma no cover
    #OrderedDict = dict
    from collections import OrderedDict
except ImportError:  # pragma no cover
    from ordereddict import OrderedDict

try:  # pragma no cover
    _basestring = basestring
except NameError:  # pragma no cover
    _basestring = str
try:  # pragma no cover
    _unicode = unicode
except NameError:  # pragma no cover
    _unicode = str

from pyutilib.misc.pyyaml_util import yaml_eval

__author__ = 'Martin Blech'
__version__ = '0.2'
__license__ = 'MIT'


class ParsingInterrupted(Exception):
    pass


class _DictSAXHandler:

    def __init__(self,
                 item_depth=0,
                 item_callback=lambda *args: True,
                 xml_attribs=True,
                 attr_prefix='@',
                 cdata_key='#text',
                 force_cdata=False,
                 cdata_separator=''):
        self.path = []
        self.stack = []
        self.data = None
        self.item = None
        self.item_depth = item_depth
        self.xml_attribs = xml_attribs
        self.item_callback = item_callback
        self.attr_prefix = attr_prefix
        self.cdata_key = cdata_key
        self.force_cdata = force_cdata
        self.cdata_separator = cdata_separator

    def startElement(self, name, attrs):
        self.path.append((name, attrs or None))
        if len(self.path) > self.item_depth:
            self.stack.append((self.item, self.data))
            attrs = OrderedDict((self.attr_prefix + key, yaml_eval(value))
                                for (key, value) in attrs.items())
            self.item = self.xml_attribs and attrs or None
            self.data = None

    def endElement(self, name):
        if len(self.path) == self.item_depth:
            item = self.item
            if item is None:
                item = self.data
            should_continue = self.item_callback(self.path, item)
            if not should_continue:
                raise ParsingInterrupted()
        if len(self.stack):
            item, data = self.item, self.data
            self.item, self.data = self.stack.pop()
            if data and self.force_cdata and item is None:
                item = OrderedDict()
            if item is not None:
                if data:
                    item[self.cdata_key] = data
                self.push_data(name, item)
            else:
                self.push_data(name, data)
        else:
            self.item = self.data = None
        self.path.pop()

    def characters(self, data):
        if data.strip():
            if not self.data:
                self.data = data
            else:
                self.data += self.cdata_separator + data

    def push_data(self, key, data):
        if self.item is None:
            self.item = OrderedDict()
        try:
            value = self.item[key]
            if isinstance(value, list):
                value.append(data)
            else:
                self.item[key] = [value, yaml_eval(data)]
        except KeyError:
            self.item[key] = yaml_eval(data)


def parse(xml_input, *args, **kwargs):
    """Parse the given XML input and convert it into a dictionary.

    `xml_input` can either be a `string` or a file-like object.

    If `xml_attribs` is `True`, element attributes are put in the dictionary
    among regular child elements, using `@` as a prefix to avoid collisions. If
    set to `False`, they are just ignored.

    Simple example::

        >>> doc = xmltodict.parse(\"\"\"
        ... <a prop="x">
        ...   <b>1</b>
        ...   <b>2</b>
        ... </a>
        ... \"\"\")
        >>> doc['a']['@prop']
        u'x'
        >>> doc['a']['b']
        [u'1', u'2']

    If `item_depth` is `0`, the function returns a dictionary for the root
    element (default behavior). Otherwise, it calls `item_callback` every time
    an item at the specified depth is found and returns `None` in the end
    (streaming mode).

    The callback function receives two parameters: the `path` from the document
    root to the item (name-attribs pairs), and the `item` (dict). If the
    callback's return value is false-ish, parsing will be stopped with the
    :class:`ParsingInterrupted` exception.

    Streaming example::

        >>> def handle(path, item):
        ...     print 'path:%s item:%s' % (path, item)
        ...     return True
        ... 
        >>> xmltodict.parse(\"\"\"
        ... <a prop="x">
        ...   <b>1</b>
        ...   <b>2</b>
        ... </a>\"\"\", item_depth=2, item_callback=handle)
        path:[(u'a', {u'prop': u'x'}), (u'b', None)] item:1
        path:[(u'a', {u'prop': u'x'}), (u'b', None)] item:2

    """
    handler = _DictSAXHandler(*args, **kwargs)
    parser = expat.ParserCreate()
    parser.StartElementHandler = handler.startElement
    parser.EndElementHandler = handler.endElement
    parser.CharacterDataHandler = handler.characters
    if hasattr(xml_input, 'read'):
        parser.ParseFile(xml_input)
    else:
        parser.Parse(xml_input, True)
    return handler.item



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ('start_ns', 'start_nsc')

import sys
import socket
from pyutilib.pyro.util import set_maxconnections, Pyro, using_pyro4


def start_ns(max_allowed_connections=None):
    if Pyro is not None:
        set_maxconnections(max_allowed_connections=max_allowed_connections)
        args = []
        for val in sys.argv[1:]:
            if val.startswith("-n") or val.startswith("--host"):
                args = list(sys.argv[1:])
                break
        else: # execute when no break occurs
            args = list(sys.argv[1:])
            args.append("-n")
            args.append(str(socket.gethostname()))
        Pyro.naming.main(args)
    else:
        raise ImportError("Pyro or Pyro4 is not installed")


def start_nsc():
    if Pyro is not None:
        if using_pyro4:
            import Pyro4
            import Pyro4.nsc
        else:
            import Pyro
            import Pyro.nsc
        Pyro.nsc.main(sys.argv[1:])
    else:
        raise ImportError("Pyro or Pyro4 is not installed")



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

# Q: When passing options, these values do not initialize the startup task.
#    Is this a bug?
# TODO: only set option values for variables that show up in a workflow's inputs
# TODO: add graceful management of exceptions
#       show the task tree, etc...

__all__ = ['Workflow']

import argparse
from collections import deque
from six import iterkeys

from pyutilib.workflow.task import Task, EmptyTask, NoTask
from pyutilib.misc import Options

try:
    from collections import OrderedDict
except:
    from ordereddict import OrderedDict


def _collect_parser_groups(t):
    for key in t._parser_group:
        #
        # NOTE: we are changing the properties of the group
        # instances here.  This is OK _only_ because we are
        # printing the help info and then terminating.
        #

        # FIXME: The 'parser' object is not defined in the lines below
        raise NotImplementedError("")
        #t._parser_group[key].parser = parser
        #parser.add_argument_group(t._parser_group[key])


def _set_arguments(t):
    for arg in t._parser_arg:
        args = arg[0]
        kwargs = arg[1]
        try:
            t._parser.add_argument(*args, **kwargs)
        except argparse.ArgumentError:
            pass


class Workflow(Task):

    def __init__(self, id=None, name=None, parser=None):
        Task.__init__(self, id=id, name=name, parser=None)
        self._tasks = {}
        self._start_task = EmptyTask()
        self._final_task = EmptyTask()
        self.add(self._start_task)
        self.add(self._final_task)

    def add(self, task, loadall=True):
        if self.debug:
            print("ADDING", task.id)  #pragma:nocover
        if task.id == NoTask.id:
            return
        if task.id in self._tasks:
            return
        self._tasks[task.id] = task
        if not loadall:
            return
        for name in task.inputs:
            for t in task.inputs[name].from_tasks():
                self.add(t)

            if len(task.inputs[name].from_tasks()) > 0:
                continue

            if not task.inputs[name].constant:
                #
                # Constant input values are not added to the start task
                #
                if not name in self._start_task.outputs:
                    self._start_task.outputs.declare(name)
                    self.inputs.declare(name, optional=True)
                # TODO: this is a bit of a hack...
                val = getattr(task.inputs, name).get_value()
                try:
                    setattr(task.inputs, name, getattr(self._start_task.outputs,
                                                       name))
                except ValueError:
                    # TBD: when do we get this exception?
                    pass
                getattr(self.inputs, name).set_value(val)
        for name in task.output_controls:
            for c in task.output_controls[name].output_connections:
                self.add(c.to_port.task())
        #
        for name in task.outputs:
            if len(task.outputs[name].output_connections) > 0:
                for c in task.outputs[name].output_connections:
                    self.add(c.to_port.task())
            else:
                if name in self._final_task.inputs:
                    raise ValueError(
                        "Cannot declare a workplan with multiple output values that share the same name: %s"
                        % name)
                self.outputs.declare(name)
                self._final_task.inputs.declare(name)
                setattr(self._final_task.inputs, name, task.outputs[name])

    def _call_init(self, *options, **kwds):
        Task._call_init(self, *options, **kwds)
        for i in self.inputs:
            val = self.inputs[i].get_value()
            if val is not None:
                self._start_task.outputs[i].set_value(val)
                self._start_task.outputs[i].set_ready()
        #
        # TBD: this appears to be redundant
        #
        #for key in kwds:
        #    if key not in self._start_task.outputs:
        #        raise ValueError, "Cannot specify value for option %s.  Valid option names are %s" % (key, self._start_task.outputs.keys())
        #    self._start_task.outputs[key].set_value( kwds[key] )

    def _call_fini(self, *options, **kwds):
        ans = Options()
        for key in self._final_task.inputs:
            self._final_task.inputs[key].compute_value()
            ans[key] = self._final_task.inputs[key].get_value()
            getattr(self.outputs, key).set_value(ans[key])
        for key in self.outputs:
            self.outputs[key].set_ready()
        return ans

    def set_options(self, args):
        self._dfs_([self._start_task.id], lambda t: t.set_options(args))

    def options(self):
        return self._start_task.outputs.keys()

    def print_help(self):
        parser = argparse.ArgumentParser()
        self._dfs_([self._start_task.id], _collect_parser_groups)
        parser.print_help()

    def set_arguments(self, parser=None):
        if parser is None:
            parser = self._parser
        self._dfs_([self._start_task.id], _set_arguments)

    def reset(self):
        return self._dfs_([self._start_task.id], lambda t: t.reset())

    def execute(self):
        #return self._dfs_([self._start_task.id], lambda t: t.__call__())
        if self.debug:  #pragma:nocover
            print(self.name, '---------------')
            print(self.name, '---------------')
            print(self.name, '   STARTING')
            print(self.name, '---------------')
            print(self.name, '---------------')
        #
        queued = set([self._start_task.id])
        queue = deque([self._start_task])
        waiting = OrderedDict()
        while len(queue) + len(waiting) > 0:
            for id in list(iterkeys(waiting)):
                t = waiting[id]
                if not t.id in queued and t.ready():
                    #
                    if self.debug:  #pragma:nocover
                        print(self.name, "WAITING: ", t.id,
                              " not queued and task ready")
                        print(self.name, "Waiting task", t.name, t.id,
                              t.ready())
                    #
                    queue.append(t)
                    queued.add(t.id)
                    del waiting[t.id]
            if len(queue) == 0:
                break
                # TBD: should we sleep and add a timelimit before raising this exception?
                #if len(waiting) == 0:
                #    break
                #print self.name, "ERROR", waiting.keys()
                #raise RuntimeError, "Workflow failed to terminate normally.  All available tasks are blocked."
            task = queue.popleft()
            #
            if self.debug:  #pragma:nocover
                print(self.name, "TASK   ", str(task))
                print(self.name, "QUEUE  ", queued)
                print(self.name, "WAITING", waiting.keys())
                print(self.name, "Executing Task " + task.name,
                      task.next_task_ids())
            #
            queued.remove(task.id)
            task()
            for t in task.next_tasks():
                if t.id in queued:
                    continue
                if t.ready():
                    #
                    if self.debug:  #pragma:nocover
                        print(self.name, "NEXT: ", t.id,
                              " not queued and task ready")
                        print(self.name, "Scheduling task", t.name, t.id,
                              t.ready())
                    #
                    queue.append(t)
                    queued.add(t.id)
                    if t.id in waiting:
                        del waiting[t.id]
                else:
                    if not t.id in waiting:
                        waiting[t.id] = t
                    #
                    if self.debug:  #pragma:nocover
                        print(self.name, "NEXT: ", t.id,
                              " not queued and task NOT ready")
                        print(self.name, "Ignoring task", t.name, t.id,
                              t.ready())
            if self.debug:  #pragma:nocover
                print(self.name, "FINAL QUEUE  ", queued)
                print(self.name, "FINAL WAITING", waiting.keys())
                print(self.name, '---------------')
                print(self.name, "    LOOP")
                print(self.name, '---------------')
        if self.debug:  #pragma:nocover
            print(self.name, '---------------')

    def __str__(self):
        return "\n".join(["Workflow %s:" % self.name] + self._dfs_(
            [self._start_task.id], lambda t: t._name()))

    def __repr__(self):
        return "Workflow %s:\n" % self.name + Task.__repr__(
            self) + '\n' + "\n".join(
                self._dfs_([self._start_task.id], lambda t: str(t)))

    def _dfs_(self, indices, fn, touched=None):
        if touched is None:
            touched = set()
        ans = []
        for i in indices:
            if i in touched:
                # With this design, this condition should never be triggered
                # TODO: verify that this is an O(n) search algorithm; I think it's
                # O(n^2)
                continue  #pragma:nocover
            ok = True
            task = self._tasks[i]
            for j in task.prev_task_ids():
                if (j == NoTask.id) or (j in touched):
                    continue
                ok = False
                break
            if not ok:
                continue
            tmp = fn(task)
            if tmp is not None:
                ans.append(tmp)
            touched.add(i)
            ans = ans + self._dfs_(task.next_task_ids(), fn, touched)
        return ans



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

import logging
import os
from inspect import currentframe, getframeinfo
from six import StringIO

import pyutilib.misc
import pyutilib.th as unittest

from pyutilib.misc.log_config import LogHandler

logger = logging.getLogger('pyutilib-testing')
filename = getframeinfo(currentframe()).filename

class TestLogging(unittest.TestCase):
    def setUp(self):
        self.stream = StringIO()

    def tearDown(self):
        logger.removeHandler(self.handler)

    def test_simple_log(self):
        # Testing positional base, configurable verbosity
        self.handler = LogHandler(
            os.path.dirname(__file__),
            stream = self.stream,
            verbosity=lambda: logger.isEnabledFor(logging.DEBUG))
        logger.addHandler(self.handler)

        logger.setLevel(logging.WARNING)
        logger.info("(info)")
        self.assertEqual(self.stream.getvalue(), "")
        logger.warn("(warn)")
        ans = "WARNING: (warn)\n"
        self.assertEqual(self.stream.getvalue(), ans)

        logger.setLevel(logging.DEBUG)
        logger.warn("(warn)")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += 'WARNING: "[base]%stest_log_config.py", %d, test_simple_log\n' \
               '    (warn)\n' % (os.path.sep, lineno,)
        self.assertEqual(self.stream.getvalue(), ans)

    def test_alternate_base(self):
        self.handler = LogHandler(
            base = 'log_config',
            stream = self.stream)
        logger.addHandler(self.handler)

        logger.setLevel(logging.WARNING)
        logger.info("(info)")
        self.assertEqual(self.stream.getvalue(), "")
        logger.warn("(warn)")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans = 'WARNING: "%s", %d, test_alternate_base\n' \
               '    (warn)\n' % (filename, lineno,)
        self.assertEqual(self.stream.getvalue(), ans)

    def test_no_base(self):
        self.handler = LogHandler(
            stream = self.stream)
        logger.addHandler(self.handler)

        logger.setLevel(logging.WARNING)
        logger.info("(info)")
        self.assertEqual(self.stream.getvalue(), "")
        logger.warn("(warn)")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans = 'WARNING: "%s", %d, test_no_base\n' \
               '    (warn)\n' % (filename, lineno,)
        self.assertEqual(self.stream.getvalue(), ans)

    def test_no_message(self):
        self.handler = LogHandler(
            os.path.dirname(__file__),
            stream = self.stream,
            verbosity=lambda: logger.isEnabledFor(logging.DEBUG))
        logger.addHandler(self.handler)

        logger.setLevel(logging.WARNING)
        logger.info("")
        self.assertEqual(self.stream.getvalue(), "")

        logger.warn("")
        ans = "WARNING\n"
        self.assertEqual(self.stream.getvalue(), ans)

        logger.setLevel(logging.DEBUG)
        logger.warn("")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += 'WARNING: "[base]%stest_log_config.py", %d, test_no_message\n' \
               % (os.path.sep, lineno,)
        self.assertEqual(self.stream.getvalue(), ans)

    def test_numbered_level(self):
        testname ='test_numbered_level'
        self.handler = LogHandler(
            os.path.dirname(__file__),
            stream = self.stream,
            verbosity=lambda: logger.isEnabledFor(logging.DEBUG))
        logger.addHandler(self.handler)

        logger.setLevel(logging.WARNING)
        logger.log(45, "(hi)")
        ans = "Level 45: (hi)\n"
        self.assertEqual(self.stream.getvalue(), ans)

        logger.log(45, "")
        ans += "Level 45\n"
        self.assertEqual(self.stream.getvalue(), ans)

        logger.setLevel(logging.DEBUG)
        logger.log(45, "(hi)")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += 'Level 45: "[base]%stest_log_config.py", %d, %s\n' \
               '    (hi)\n' % (os.path.sep, lineno, testname)
        self.assertEqual(self.stream.getvalue(), ans)

        logger.log(45, "")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += 'Level 45: "[base]%stest_log_config.py", %d, %s\n' \
               % (os.path.sep, lineno, testname)
        self.assertEqual(self.stream.getvalue(), ans)

    def test_long_messages(self):
        self.handler = LogHandler(
            os.path.dirname(__file__),
            stream = self.stream,
            verbosity=lambda: logger.isEnabledFor(logging.DEBUG))
        logger.addHandler(self.handler)

        msg = ("This is a long message\n\n"
               "With some kind of internal formatting\n"
               "    - including a bulleted list\n"
               "    - list 2  ")
        logger.setLevel(logging.WARNING)
        logger.warn(msg)
        ans = ( "WARNING: This is a long message\n\n"
                "    With some kind of internal formatting\n"
                "        - including a bulleted list\n"
                "        - list 2\n" )
        self.assertEqual(self.stream.getvalue(), ans)

        logger.setLevel(logging.DEBUG)
        logger.info(msg)
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += ( 'INFO: "[base]%stest_log_config.py", %d, test_long_messages\n'
                 "    This is a long message\n\n"
                 "    With some kind of internal formatting\n"
                 "        - including a bulleted list\n"
                 "        - list 2\n" % (os.path.sep, lineno,))
        self.assertEqual(self.stream.getvalue(), ans)

        # test trailing newline
        msg += "\n"
        logger.setLevel(logging.WARNING)
        logger.warn(msg)
        ans += ( "WARNING: This is a long message\n\n"
                "    With some kind of internal formatting\n"
                "        - including a bulleted list\n"
                "        - list 2\n" )
        self.assertEqual(self.stream.getvalue(), ans)

        logger.setLevel(logging.DEBUG)
        logger.info(msg)
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += ( 'INFO: "[base]%stest_log_config.py", %d, test_long_messages\n'
                 "    This is a long message\n\n"
                 "    With some kind of internal formatting\n"
                 "        - including a bulleted list\n"
                 "        - list 2\n" % (os.path.sep, lineno,))
        self.assertEqual(self.stream.getvalue(), ans)

        # test initial and final blank lines
        msg = "\n" + msg + "\n\n"
        logger.setLevel(logging.WARNING)
        logger.warn(msg)
        ans += ( "WARNING: This is a long message\n\n"
                "    With some kind of internal formatting\n"
                "        - including a bulleted list\n"
                "        - list 2\n" )
        self.assertEqual(self.stream.getvalue(), ans)

        logger.setLevel(logging.DEBUG)
        logger.info(msg)
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += ( 'INFO: "[base]%stest_log_config.py", %d, test_long_messages\n'
                 "    This is a long message\n\n"
                 "    With some kind of internal formatting\n"
                 "        - including a bulleted list\n"
                 "        - list 2\n" % (os.path.sep, lineno,))
        self.assertEqual(self.stream.getvalue(), ans)



__all__ = ['FileResource']

import os.path
from pyutilib.workflow import resource


class FileResource(resource.Resource):

    def __init__(self, name=None):
        resource.Resource.__init__(self)
        self.filename = name
        if name is None:
            self.name = "File" + self.id
        else:
            self.name = "File_" + os.path.basename(name)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['ply_init', 't_newline', 't_ignore', 't_COMMENT', '_find_column',
           'p_error']

#
# Utility functions that are used with PLY
#


def ply_init(data):
    global _parsedata
    _parsedata = data
    global _init
    _init = True


def t_newline(t):
    r'[\n]+'
    global _init
    if _init:
        t.lexer.lineno = len(t.value)
        _init = False
    else:
        t.lexer.lineno += len(t.value)

# Ignore space and tab
t_ignore = " \t\r"


# Discard comments
def t_COMMENT(t):
    r'\#[^\n]*'
    pass
# Tokens in comments are discarded.


#
# Compute column.
#     input is the input text string
#     token is a token instance
#
def _find_column(input, token):
    i = token.lexpos
    while i > 0:
        if input[i] == '\n': break
        i -= 1
    column = (token.lexpos - i) + 1
    return column


def p_error(p):
    if p is None:
        tmp = "Syntax error at end of file."
    else:
        tmp = "Syntax error at token "
        if p.type is "":
            tmp = tmp + "''"
        else:
            tmp = tmp + str(p.type)
        tmp = tmp + " with value '" + str(p.value) + "'"
        tmp = tmp + " in line " + str(p.lineno)
        tmp = tmp + " at column " + str(_find_column(_parsedata, p))
    raise IOError(tmp)



#!/usr/bin/env python

import sys
import os
from os.path import abspath, basename, dirname, exists, join
import subprocess

usage = """usage: lbin [-f][-v] <command> [...]

This script recurses up the current path, looking for 'bin' and
python/bin directories.  If found, these directories are prepended to
the PATH environment path.
    -f : [find_only] show results from 'which <command>' after PATH
         expansion but do not run <command>
    -v : [verbose] show results from 'which <command>' after PATH expansion
"""


def lbin(args):
    verbose = False
    find_only = False
    if args and args[0] in ['-v', '-f']:
        if args[0] == '-v':
            verbose = True
        elif args[0] == '-f':
            find_only = True
        args.pop(0)

        if args and args[0] in ['-v', '-f']:
            if args[0] == '-v':
                verbose = True
            elif args[0] == '-f':
                find_only = True
            args.pop(0)

    if len(args) == 0:
        print(usage)
        return 1

    try:
        # On unix, the shell's reported PWD may be different from the '.'
        # inode's absolute path (e.g., when the cwd was reached through
        # symbolic links.)  This is a trivial way to defer to the shell's
        # notion of the cwd and quietly fall back on Python's getcwd() for
        # non-unix environments.
        curr = os.environ['PWD']
    except:
        curr = abspath(os.getcwd())
    dirs = []
    while os.sep in curr:
        if exists(join(curr, "python")):
            dirs.append(join(curr, "python", "bin"))
        if exists(join(curr, "bin")):
            dirs.append(join(curr, "bin"))
        if basename(curr) == "":
            break
        curr = dirname(curr)

    dirs.append(os.environ["PATH"])
    os.environ["PATH"] = os.pathsep.join(dirs)
    #print os.environ["PATH"]
    try:
        if find_only:
            if verbose:
                return subprocess.call(['which', '-a', args[0]])
            else:
                return subprocess.call(['which', args[0]])
        if verbose:
            print("Path search found the following instances of %s:" % args[0])
            x = subprocess.call(['which', args[0]])
            print("")
        os.execvp(args[0], args)
        #return subprocess.call(args)
    except OSError:
        err = sys.exc_info()[1]
        print("ERROR executing command '%s': %s" % (' '.join(args), str(err)))
        return err.errno


    # The [console_scripts] entry point requires a function that takes no
    # arguments
def main():
    sys.exit(lbin(sys.argv[1:]))


if __name__ == '__main__':
    main()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

import sys

__all__ = ['ConfigurationError', 'ApplicationError', 'BadDebuggingValue']


class WindowsError_def(Exception):
    """
    An exception used there is an error configuring a package.
    """

    def __init__(self, *args, **kargs):
        Exception.__init__(self, *args, **kargs)  #pragma:nocover


if (sys.platform[0:3] != "win"):
    WindowsError = WindowsError_def
else:
    WindowsError = WindowsError
__all__.append("WindowsError")


class ConfigurationError(Exception):
    """
    An exception used there is an error configuring a package.
    """

    def __init__(self, *args, **kargs):
        Exception.__init__(self, *args, **kargs)  #pragma:nocover


class ApplicationError(Exception):
    """
    An exception used when an external application generates an error.
    """

    def __init__(self, *args, **kargs):
        Exception.__init__(self, *args, **kargs)  #pragma:nocover


class BadDebuggingValue(Exception):
    """
    An exception used when a bad debugging value is used.
    """

    def __init__(self, *args, **kargs):
        Exception.__init__(self, *args, **kargs)  #pragma:nocover



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['get_xml_text', 'escape', 'compare_xml_files']

from xml.dom import Node
import sys


def get_xml_text(node):
    nodetext = ""
    for child in node.childNodes:
        if child.nodeType == Node.TEXT_NODE:
            nodetext = nodetext + child.nodeValue
    nodetext = str(nodetext)
    return nodetext.strip()


if sys.version_info < (3, 0):
    _identitymap = "".join([chr(n) for n in range(256)])
    _delchars = _identitymap[:9] + chr(11) + chr(12) + _identitymap[
        14:31] + chr(124)

    def escape(s):
        """Replace special characters '&', "'", '<', '>' and '"' by XML entities."""
        s = s.replace("&", "&amp;")  # Must be done first!
        s = s.replace("<", "&lt;")
        s = s.replace(">", "&gt;")
        s = s.translate(_identitymap, _delchars)
        return s

else:

    _tmap = {}
    for n in range(256):
        if n < 9 or (n >= 14 and n < 31) or n in (11, 12, 124):
            _tmap[chr(n)] = None

    def escape(s):
        """Replace special characters '&', "'", '<', '>' and '"' by XML entities."""
        s = s.replace("&", "&amp;")  # Must be done first!
        s = s.replace("<", "&lt;")
        s = s.replace(">", "&gt;")
        s = s.translate(_tmap)
        return s


def compare_xml_files(baseline_fname,
                      output_fname,
                      tolerance=0.0,
                      baseline_begin='',
                      baseline_end='',
                      output_begin='',
                      output_end=None,
                      exact=True):
    from pyutilib.misc.pyyaml_util import extract_subtext, compare_repn
    from pyutilib.misc.xmltodict import parse
    #
    INPUT = open(baseline_fname, 'r')
    baseline = extract_subtext(
        INPUT, begin_str=baseline_begin, end_str=baseline_end)
    INPUT.close()
    #
    INPUT = open(output_fname, 'r')
    output = extract_subtext(INPUT, begin_str=output_begin, end_str=output_end)
    INPUT.close()
    #
    baseline_repn = parse(baseline)
    output_repn = parse(output)
    return compare_repn(
        baseline_repn, output_repn, tolerance=tolerance, exact=exact)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""
The outline of the PyUtilib Component Architecture (PCA) is adapted from Trac
(see the trac.core module).  This framework generalizes the Trac by supporting
multi-environment management of components, as well as non-singleton plugins.

This package provides a stand-alone module that defines all of the core
aspects of the PCA.  Related Python packages define extensions of this
framework that support current component-based applications.

NOTE: The PCA does not rely on any other part of PyUtilib.  Consequently,
this package can be independently used in other projects.
"""

from pyutilib.component.core.core import Plugin, implements, Interface, CreatePluginFactory, \
           PluginMeta, alias, ExtensionPoint, SingletonPlugin, \
           PluginFactory, PluginError, PluginGlobals, with_metaclass, \
           IPluginLoader, IPluginLoadPath, IIgnorePluginWhenLoading, \
           IOptionDataProvider, PluginEnvironment

PluginGlobals.add_env("pca")

#
# This declaration is here because this is a convenient place where
# all symbols in this module have been defined.
#


class IgnorePluginPlugins(SingletonPlugin):
    """Ignore plugins from the pyutilib.component module"""

    implements(IIgnorePluginWhenLoading)

    def ignore(self, name):
        return name in globals()

#
# Import the 'pyutilib.component' plugins
#
try:
    import pkg_resources
    #
    # Load modules associated with Plugins that are defined in
    # EGG files.
    #
    for entrypoint in pkg_resources.iter_entry_points('pyutilib.component'):
        plugin_class = entrypoint.load()
        # print "Loading plugins... (%s)" % entrypoint
except ImportError:
    pass
except Exception:
    import sys
    err = sys.exc_info()[1]
    from sys import stderr as SE
    SE.write("Error loading 'pyutilib.component' entry points: '%s'\n" % err)

#
# Remove the "pca" environment as the default
#
PluginGlobals.pop_env()



import pyutilib.component.core
import sys
import os
import logging

currdir = sys.argv[-2] + os.sep

pyutilib.component.core.PluginGlobals.add_env(
    pyutilib.component.core.PluginEnvironment("testing"))
service = pyutilib.component.core.PluginFactory(
    "EggLoader", namespace='project1', env='pca')
pyutilib.misc.setup_redirect(currdir + "egg1.out")
if service is None:
    print(
        "Cannot test the PyUtilib EggLoader Plugin on this system because the pkg_resources package is not available.")
    sys.exit(1)
#
#logging.basicConfig(level=logging.DEBUG)
#
pyutilib.component.core.PluginGlobals.get_env().load_services(
    path=currdir + "eggs1")
#
if sys.argv[-1] == 'json':
    pyutilib.component.core.PluginGlobals.pprint(json=True)
else:
    pyutilib.component.core.PluginGlobals.pprint()
pyutilib.misc.reset_redirect()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""Plugins that contain options that support configuration of their services."""

__all__ = ['ManagedPlugin', 'ManagedSingletonPlugin']

from pyutilib.component.core import Plugin, SingletonPlugin
from pyutilib.component.config.options import declare_option, BoolOption


class ManagedPlugin(Plugin):
    """A plugin that has an option supports configuration of this service."""

    def __init__(self, **kwds):
        Plugin.__init__(self, **kwds)
        #super(ManagedPlugin,self).__init__(**kwds)
        declare_option(
            name=self.name,
            section="Services",
            local_name="enable",
            default=self._enable,
            cls=BoolOption,
            doc="Option that controls behavior of service %s." % self.name)

    def __del__(self):
        # JPW: There has to be a better way to determine/cache the name of the option created in the 
        #      constructor call. For example, declare_option might return the name of the created attribute.
        #      In the mean time, I have hard-coded the (known) name.

        # clean up the attached BoolOption, which is a plugin - otherwise, a memory leak will result.
        self._enable.deactivate()


class ManagedSingletonPlugin(SingletonPlugin):
    """A singleton plugin that has an option supports configuration of this service."""

    def __init__(self, **kwds):
        Plugin.__init__(self, **kwds)
        #super(ManagedSingletonPlugin,self).__init__(**kwds)
        declare_option(
            name=self.name,
            section="Services",
            local_name="enable",
            default=self._enable,
            cls=BoolOption,
            doc="Option that controls behavior of service %s." % self.name)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""
The pyutilib.component.config package includes utilities to configure
the PyUtilib Component Architecture.  This includes facilities for using
configuration files, controlling logging, and specifying component options.
"""

from pyutilib.component.core import PluginGlobals
PluginGlobals.add_env("pca")

from pyutilib.component.config.env_config import EnvironmentConfig
from pyutilib.component.config.options import ExecutableOption, declare_option, Option
from pyutilib.component.config.managed_plugin import ManagedPlugin, ManagedSingletonPlugin
from pyutilib.component.config.configuration import Configuration, ConfigurationError
from pyutilib.component.config.logging_config import LoggingConfig
from pyutilib.component.config.tempfiles import ITempfileManager, TempfileManagerPlugin, TempfileManager
import pyutilib.component.config.plugin_ConfigParser

PluginGlobals.pop_env()



#
# Unit Tests for util/singleton
#
#

import os
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep
import pyutilib.th as unittest
import pyutilib.misc


class A(pyutilib.misc.MonoState):

    def __init__(self):
        self.state = True


class B(pyutilib.misc.Singleton):

    def __init__(self):
        self.state = True


class SingletonDebug(unittest.TestCase):

    def test_A(self):
        # Verify that MonoState generates one global state
        a1 = A()
        a2 = A()
        self.assertNotEqual(a1, a2)
        self.assertEqual(a1.__dict__, a2.__dict__)

    def test_B(self):
        # Verify that Singleton generates one instance
        b1 = B()
        b2 = B()
        self.assertEqual(b1, b2)


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

import pyutilib.misc.GlobalData
from pyutilib.misc.archivereader import ArchiveReaderFactory, ArchiveReader,\
           ZipArchiveReader, TarArchiveReader, DirArchiveReader, FileArchiveReader,\
           GzipFileArchiveReader, BZ2FileArchiveReader
from pyutilib.misc.comparison import compare_file_with_numeric_values, compare_file, compare_large_file
from pyutilib.misc.cross import cross, cross_iter, flattened_cross_iter
from pyutilib.misc.dict_with_default import SparseMapping
from pyutilib.misc.factory import Factory
from pyutilib.misc.format_io import format_float, format_io
from pyutilib.misc.gc_manager import PauseGC
from pyutilib.misc.generator import coroutine
from pyutilib.misc.import_file import import_file, run_file
from pyutilib.misc.indent_io import StreamIndenter
from pyutilib.misc.log_config import LogHandler
from pyutilib.misc.method import add_method, add_method_by_name
from pyutilib.misc.misc import deprecated, tostr, flatten, flatten_list, recursive_flatten_tuple, flatten_tuple, handleRemoveReadonly, rmtree, quote_split, traceit, tuplize, find_files, search_file, sort_index, count_lines, Bunch, Container, Options, create_hardlink, executable_extension
from pyutilib.misc.pyyaml_util import yaml_fix, json_fix, load_yaml, load_json, extract_subtext, compare_repn, compare_strings, compare_yaml_files, compare_json_files, simple_yaml_parser
from pyutilib.misc.redirect_io import capture_output, setup_redirect, reset_redirect
from pyutilib.misc.singleton import Singleton, MonoState
from pyutilib.misc.tee_io import TeeStream, ConsoleBuffer
from pyutilib.misc.xml_utils import get_xml_text, escape, compare_xml_files
from pyutilib.misc.visitor import SimpleVisitor, ValueVisitor



# Imports
import sys
import os
from os.path import dirname, abspath, abspath

import pyutilib.th as unittest

currdir = dirname(abspath(__file__)) + os.sep
datadir = os.sep.join([dirname(dirname(dirname(dirname(abspath(__file__))))),
                       'doc', 'workflow', 'examples']) + os.sep

sys.path.insert(0, datadir)
try:
    from test_example import *
finally:
    sys.path.remove(datadir)

# Execute the tests
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# A script that verifies that suitable copyright statements are 
# defined in files.
#

import sys
import os
import re

suffixes = ['.c', '.cc', '.h', '.H', '.cpp', '.py']


def recurse(dir):
    for root, dirnames, filenames in os.walk(dir):
        for filename in filenames:
            for suffix in suffixes:
                if filename.endswith(suffix):
                    yield os.path.join(root, filename)


def match(filename, pat):
    INPUT = open(filename, 'r')
    fstring = INPUT.read()
    INPUT.close()
    if pat.search(fstring):
        return True
    return False


def main():
    nfiles = 0
    badfiles = []
    if sys.argv[1] == '-c':
        cfile = sys.argv[2]
        files = sys.argv[3:]
        INPUT = open(cfile, 'r')
        cstring = INPUT.read()
        INPUT.close()
        pat = re.compile(cstring)
    else:
        cfile = None
        cstring = None
        pat = re.compile('Copyright')
        files = sys.argv[1:]
    for dir_ in files:
        for filename in recurse(dir_):
            nfiles += 1
            if not match(filename, pat):
                badfiles.append(filename)

    print("Total number of files missing copyright: " + str(len(badfiles)))
    print("Total number of files checked:           " + str(nfiles))
    if len(badfiles) > 0:
        print("")
        print("Bad Files")
        print("-" * 40)
        for filename in badfiles:
            print(filename)


if __name__ == '__main__':
    main()



#
# Unit Tests for json (in json_utils)
#
#

import os
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep
import pyutilib.th as unittest
import pyutilib.misc

try:
    import json
    json_available = True
except ImportError:
    json_available = False


class Test(unittest.TestCase):

    def setUp(self):
        if not json_available:
            self.skipTest("Cannot execute test because JSON is not available.")

    def test1(self):
        # Verify that data 1 is not equal to data 2
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_json_files,
            currdir + 'jsondata1.jsn',
            currdir + 'jsondata2.jsn',
            exact=True)

    def test2(self):
        # Verify that data 1 is a subset of data 2
        pyutilib.misc.compare_json_files(
            currdir + 'jsondata1.jsn', currdir + 'jsondata2.jsn', exact=False)

    def test3(self):
        # Verify that data 2 is not a subset of data 1
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_json_files,
            currdir + 'jsondata2.jsn',
            currdir + 'jsondata1.jsn',
            exact=True)

    def test4(self):
        # Verify check to confirm that YAML data has a different type
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_json_files,
            currdir + 'jsondata1.jsn',
            currdir + 'jsondata3.jsn',
            exact=True)

    def test5(self):
        # Verify that data 1 is not equal to data 1
        pyutilib.misc.compare_json_files(
            currdir + 'jsondata1.jsn', currdir + 'jsondata1.jsn', exact=True)

    def test6(self):
        # Verify that data 1 is a subset of data 4
        pyutilib.misc.compare_json_files(
            currdir + 'jsondata1.jsn',
            currdir + 'jsondata4.txt',
            output_begin='BEGIN',
            output_end='END',
            exact=False)

    def test7(self):
        # Verify that data 6 is a subset of data 5
        pyutilib.misc.compare_json_files(
            currdir + 'jsondata6.jsn', currdir + 'jsondata5.jsn', exact=False)

    def test8(self):
        # Verify that data 5 is not a subset of data 6
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_json_files,
            currdir + 'jsondata5.jsn',
            currdir + 'jsondata6.jsn',
            exact=False)

    def test9(self):
        # Verify that data 7 is not a subset of data 5
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_json_files,
            currdir + 'jsondata7.jsn',
            currdir + 'jsondata5.jsn',
            exact=False)

    def test10(self):
        # Verify that data 6 is not equal to data 5
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_json_files,
            currdir + 'jsondata6.jsn',
            currdir + 'jsondata5.jsn',
            exact=True)

    def test11(self):
        # Verify that data 8 is not equal to data 5
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_json_files,
            currdir + 'jsondata8.jsn',
            currdir + 'jsondata5.jsn',
            exact=True)

    def test12(self):
        # Verify that data 9 is not a subset of data 1
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_json_files,
            currdir + 'jsondata9.jsn',
            currdir + 'jsondata1.jsn',
            exact=True)

    def test13(self):
        # Verify that data 10 is not a subset of data 1
        self.assertRaises(
            ValueError,
            pyutilib.misc.compare_json_files,
            currdir + 'jsondata10.jsn',
            currdir + 'jsondata1.jsn',
            exact=True)

    def test14(self):
        # Verify that data 10 is a subset of data 1 with a loose tolerance
        pyutilib.misc.compare_json_files(
            currdir + 'jsondata10.jsn',
            currdir + 'jsondata1.jsn',
            tolerance=1.0,
            exact=True)

    def test15a(self):
        # Verify that compressed files can be compared
        pyutilib.misc.compare_json_files(
            currdir + 'jsondata2.jsn', currdir + 'jsondata2.jsn.gz', exact=True)

    def test15b(self):
        # Verify that compressed files can be compared
        pyutilib.misc.compare_json_files(
            currdir + 'jsondata2.jsn.gz', currdir + 'jsondata2.jsn', exact=True)


if __name__ == "__main__":
    unittest.main()



#
# Unit Tests for component/core
#

import os
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep

import pyutilib.th as unittest
import pyutilib.component.core


#
# This class is declared to facilitate coverage
#
class DummyPlugin(pyutilib.component.core.Plugin):

    pyutilib.component.core.implements(pyutilib.component.core.IPluginLoadPath)
    pyutilib.component.core.implements(pyutilib.component.core.IPluginLoader)

    def get_load_path(self):
        return []

    def load(self, x, y, z, w):
        pass


class TestLoader(unittest.TestCase):

    def setUp(self):
        pyutilib.component.core.PluginGlobals.add_env("testing.loader")
        DummyPlugin()

    def tearDown(self):
        pyutilib.component.core.PluginGlobals.remove_env("testing.loader")

    def test_load1(self):
        pyutilib.component.core.PluginGlobals.get_env().load_services(
            path=currdir + "plugins1", auto_disable=True, name_re="^$")

    def test_load2(self):
        pyutilib.component.core.PluginGlobals.get_env().load_services(
            path=[currdir + "plugins1", currdir + "plugins2"],
            auto_disable="^$",
            name_re="^$")

    def test_load3(self):
        try:
            pyutilib.component.core.PluginGlobals.get_env().load_services(
                path={})
            self.fail("expect error")
        except pyutilib.component.core.PluginError:
            pass

    def test_load4(self):
        pyutilib.component.core.PluginGlobals.load_services(
            auto_disable="", name_re="^$")

    def test_load5(self):
        pyutilib.component.core.PluginGlobals.get_env().load_services(
            path=[currdir + "plugins1", currdir + "plugins2"],
            auto_disable=False,
            name_re="^$")

    def test_load6(self):
        pyutilib.component.core.PluginGlobals.get_env().load_services(
            path=[currdir + "plugins1", currdir + "plugins2"],
            auto_disable=False,
            name_re=True)


if __name__ == "__main__":
    unittest.main()
